/*
Dejfif: extract jpeg images from any binary container that encapsulates
	them among other data.

Copyright (C) 2005 Andrea Sottoriva 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, 
USA.
*/

#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

unsigned char SOI_marker[] =  { 0xff, 0xd8 };
unsigned char EOI_marker[] =  { 0xff, 0xd9 };
unsigned char JFIF_marker[] = { 0xff, 0xe0 };
unsigned char QT_marker[] =   { 0xff, 0xdb };
unsigned char HT_marker[] =   { 0xff, 0xc4 };
unsigned char SOF_marker[] =  { 0xff, 0xc0 };
unsigned char SOS_marker[] =  { 0xff, 0xda };
unsigned char JFIF_id[] =     { 0x4a, 0x46, 0x49, 0x46, 0x00 }; /* JFIF */

int main(int argc, char **argv)
{
  int f, newf, length, i, num = 0;
  unsigned char *img, str[128];
  unsigned int quant = 0, huff = 0, SOF = 0, SOS = 0;

  if(argc < 2) {
    fprintf(stderr, "Usage : ./dejfif [filename]\n");
    exit(-1);
  }

  if((f = open(argv[1], O_RDONLY)) < 0) {
    perror("open error");
    exit(-1);
  }

  if((length = lseek(f, 0, SEEK_END)) < 0) {
    perror("lseek error");
    exit(-1);
  }

  printf("File length = %d\n", length);

  lseek(f, 0, SEEK_SET);

  if((img = mmap(0, length, PROT_READ, MAP_PRIVATE, f, 0)) == NULL) {
    perror("mmap error");
    exit(-1);
  }

  for(i = 0; i < length; ++i) {
    if(img[i] == 0xff && memcmp(&img[i], SOI_marker, 2) == 0) {
      if(memcmp(&img[i + 2], JFIF_marker, 2) != 0) {
	fprintf(stderr, "No JFIF marker found, not a JFIF format\n");
	i += 4;
	continue;
      }

      if(memcmp(&img[i + 6], JFIF_id, 5) != 0) {
	fprintf(stderr, "Bad JFIF format, skip\n");
	i += 11;
	continue;
      }

      sprintf(str, "image_%03d.jpg", num);
      num++;

      fprintf(stdout, "JFIF header found at byte %d, image = %s\n", i, str);

      if((newf = open(str, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {
	perror("open new file error");
	exit(-1);
      }

      while(i < length) {

	if(write(newf, &img[i], 1) < 0) {
	  perror("write error");
	  exit(-1);
	}

	if(memcmp(&img[i], QT_marker, 2) == 0) quant = 1;
	//if(memcmp(&img[i], HT_marker, 2) == 0) huff = 1;
	if(memcmp(&img[i], SOF_marker, 2) == 0) SOF = 1;
	if(memcmp(&img[i], SOS_marker, 2) == 0) SOS = 1;
	++i;

	if((memcmp(&img[i], EOI_marker, 2) == 0) && (quant & SOF & SOS)) {
	  if(i >= length - 3 || (memcmp(&img[i + 2], SOI_marker, 2) == 0))
	    break;
	}
      }

      write(newf, &img[i], 2);

      close(newf);
    }
  }

  munmap(img, length);
}
